(************************************************************ 
 Io Interpreter                                               
   Amalthea, io_interpreter.ml                                
   Martin Sandin (d97masa@dtek.chalmers.se)                   
 ************************************************************)
 
open Io_types;;
open Io_primitives;;
open Io_env;;


(* Debug status vars *)
let trace_flag = ref false
let appl_pos = ref ("", 0, 0)
let lambda_pos = ref ("", 0, 0)


(* Environment filter, removes everything but free variables in the rest of the
   expression from the environment. *)
let filter_en
  : en_filter -> io_env -> io_env
= fun flt ((en, ven) as ten) -> match flt with
  | EFall -> ten
  | EFlimit fltis ->
      if !trace_flag then
        Io_messages.print_enfilter_trace fltis;
      (List.filter (function (i, _) -> List.mem i fltis) en, ven)

        
(* Interpreter *)
let rec interpret : Io_types.io_cont * Io_types.io_cont list -> unit
= function 
  | (Cprim f, ps) ->
      if !trace_flag then
        Io_messages.print_primitive_trace !appl_pos;
      interpret (f ps)
  | (Cexpr (Eappl (p, i, es, flt), en, _), []) ->
      appl_pos := p;
      if !trace_flag then
        Io_messages.print_application_trace !appl_pos i;
      let new_en = filter_en flt en in
      interpret (safe_lookup en i, List.map (build_cont false new_en) es)
  | (Cexpr (Elambda (p, is, e, flt), en, _), ps) ->
      lambda_pos := p;
      let cmb = try List.combine is ps
        with Invalid_argument a ->
          raise (Io_interpreter_error (Io_mismatch_error
              ("lambda", p, (List.length is, List.length ps)))) in
      let
        en =
          (if !trace_flag then
            Io_messages.print_lambda_trace p cmb);
            let new_en = filter_en flt en in
            List.fold_left bind new_en cmb
      in
      interpret (build_cont false en e, [])
  | (Cexpr (Egetvar (p, vi, i, e, flt), en, _), ps) ->
    let nen = bind en (i, !(vlookup en vi)) in
    interpret (build_cont false nen e, [])
  | (Cexpr (Eputvar (p, vi, a, e, flt), en, _), ps) ->
    vlookup en vi := build_cont false en a;
    interpret (build_cont false en e, [])
  | (Cterminate,ps) -> ()
  | (Cexpr (Eappl (p, _, _, _), _, _), ps) ->
      raise (Io_interpreter_error
        (Io_mismatch_error ("appl", p, (0, List.length ps))))
  | (_,_) ->
      raise Amalathea_AST_error


(* Interprets and catches Io error exceptions and prints them. *)
let catch_interpreter_errors
= fun f c ->
  try
    f c
  with
    | Io_interpreter_error (Io_mismatch_error (form, p, (exp ,got))) ->
        raise (Io_error (Io_messages.parameter_mismatch_error
          !appl_pos (form, p, (exp ,got))));
    | Io_interpreter_error (Io_undefined_error i) ->
        raise (Io_error (Io_messages.id_undefined_error !appl_pos i))
    | Io_prim_error m ->
        raise (Io_error (Io_messages.primitive_error m))

let interpret_expr_in_world : io_world * io_expr -> unit
= function ((_ , en, _), ex) ->
  interpret (Io_env.build_cont true en (snd (Io_env.massage_expr ex)), [])

